/*
 * AES.java
 * File revision 0
 * Hardwired guts of AES algorithm.
 *
 * Contains algorithms from the AES (Rijndael) reference implementation.
 * http://csrc.nist.gov/encryption/aes/
 */

class AES {
	Boxes boxes = new Boxes();

	void KeyAddition(short data[][], short rk[][], short BC) {
		/*
		 * XOR corresponding text input and round key input bytes.
		 */

		for(int i = 0; i < BC; ++i) {
			data[0][i] ^= rk[0][i];
			data[1][i] ^= rk[1][i];
			data[2][i] ^= rk[2][i];
			data[3][i] ^= rk[3][i];
		}

	}

	private void shiftArray(short data[], int width, int distance) {
		/*
		 * Shift an array by some distance.
		 * Won't work for distance == 0.
		 */

		int i = 0;
		short temp = data[0];

		while(i < width - 1) {
			data[i] = data[++i];
		}
		data[i] = temp;

		if(--distance > 0)
			shiftArray(data, width, distance);

	}

	void ShiftRow(short data[][], short direc, short BC) {
		/*
		 * Row 0 remains unchanged.
		 * The other three rows are shifted as per Boxes.shifts.
		 */
		int SC = ((BC - 4) >> 1);

		// data[0] unchanged
		shiftArray(data[1], BC, boxes.shifts[SC][1][direc] % BC);
		shiftArray(data[2], BC, boxes.shifts[SC][2][direc] % BC);
		shiftArray(data[3], BC, boxes.shifts[SC][3][direc] % BC);

	}

	void Substitution(short data[][], byte box[], short BC) {
		/*
		 * Replace every byte of the input by the byte at that place
		 * in the nonlinear S-box.
		 */

		for(int i = 0; i < BC; ++i) {
			data[0][i] = boxes.getElement(box, data[0][i]);
			data[1][i] = boxes.getElement(box, data[1][i]);
			data[2][i] = boxes.getElement(box, data[2][i]);
			data[3][i] = boxes.getElement(box, data[3][i]);
		}

	}

	void MixColumn(short data[][], short BC) {
		/*
		 * Mix the four bytes of every column in a linear way.
		 */
		short temp[][] = new short[4][BC];
		int i, j;

		for(i = 0; i < 4; ++i)
			for(j = 0; j < BC; ++j)
				temp[i][j] = (short)(boxes.mul((short)0x2, data[(i + 0) % 4][j]) ^
						     boxes.mul((short)0x3, data[(i + 1) % 4][j]) ^
						     boxes.mul((short)0x1, data[(i + 2) % 4][j]) ^
						     boxes.mul((short)0x1, data[(i + 3) % 4][j]));

		for(i = 0; i < BC; ++i) {
			data[0][i] = temp[0][i];
			data[1][i] = temp[1][i];
			data[2][i] = temp[2][i];
			data[3][i] = temp[3][i];
		}

	}

	void InvMixColumn(short data[][], short BC) {
		/*
		 * Mix the four bytes of every column in a linear way.
		 * This is the opposite operation of Mixcolumn.
		 */
		short temp[][] = new short[4][BC];
		int i, j;

		for(i = 0; i < 4; ++i)
			for(j = 0; j < BC; ++j)
				temp[i][j] = (short)(boxes.mul((short)0xe, data[(i + 0) % 4][j]) ^
						     boxes.mul((short)0xb, data[(i + 1) % 4][j]) ^
						     boxes.mul((short)0xd, data[(i + 2) % 4][j]) ^
						     boxes.mul((short)0x9, data[(i + 3) % 4][j]));

		for(i = 0; i < BC; ++i) {
			data[0][i] = temp[0][i];
			data[1][i] = temp[1][i];
			data[2][i] = temp[2][i];
			data[3][i] = temp[3][i];
		}

	}

	int KeySchedule(short k[][], short W[][][], short KC, short BC, short ROUNDS) {
		/*
		 * Calculate the necessary round keys.
		 *
		 * This key scheduling algorithm is a modified form of the
		 * Rijndael reference implementation's key scheduling alg.
		 */
		int i, j, t = 0, rconpointer = 0;
		short tk[][] = new short[4][KC];   

		for(j = 0; j < KC; ++j)
			for(i = 0; i < 4; ++i)
				tk[i][j] = k[i][j];

		/*
		 * Copy values into the round 0 key array.
		 */
		for(j = 0; (j < KC) && (t < (ROUNDS + 1) * BC); ++j, ++t)
			for(i = 0; i < 4; ++i)
				W[t / BC][i][t % BC] = tk[i][j];

		/*
		 * Loop until enough round key material has been computed.
		 */
		while(t < (ROUNDS + 1) * BC) {
			/*
			 * Calculate new values.
			 */
			for(i = 0; i < 4; ++i)
				tk[i][0] ^= boxes.getElement(boxes.S, tk[(i + 1) % 4][KC - 1]);
			tk[0][0] ^= boxes.rcon[rconpointer++];

			if (KC != 8) {
				for(j = 1; j < KC; ++j)
					for(i = 0; i < 4; ++i)
						tk[i][j] ^= tk[i][j - 1];
			} else {
				for(j = 1; j < KC / 2; ++j)
					for(i = 0; i < 4; ++i)
						tk[i][j] ^= tk[i][j - 1];

				for(i = 0; i < 4; ++i)
					tk[i][KC / 2] ^= boxes.getElement(boxes.S, tk[i][KC / 2 - 1]);

				for(j = KC / 2 + 1; j < KC; ++j)
					for(i = 0; i < 4; ++i) tk[i][j] ^= tk[i][j - 1];
			}

			/*
			 * Copy values into round key array.
			 */
			for(j = 0; (j < KC) && (t < (ROUNDS + 1) * BC); ++j, ++t)
				for(i = 0; i < 4; ++i)
					W[t / BC][i][t % BC] = tk[i][j];
		}

		return(0);

	}

}
